# Copyright 2021-2025 Kolena Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
import json

from kolena._api.v1.repository import CreateRepositoryRequest
from kolena._api.v1.repository import Path
from kolena._utils import krequests


def create(repository: str) -> None:
    response = krequests.post(
        endpoint_path=Path.CREATE.value,
        data=json.dumps(dataclasses.asdict(CreateRepositoryRequest(repository=repository))),
    )
    krequests.raise_for_status(response)
